from datetime import datetime
from mongoengine import (connect, ValidationError)
from nose.tools import (assert_true, assert_false, assert_equal, assert_raises)
from qiprofile_rest_client.helpers import database
from qiprofile_rest_client.model.subject import Subject
from qiprofile_rest_client.model.encounter import Encounter
from qiprofile_rest_client.model.imaging import (
    Session, Scan, ScanProtocol, Registration, RegistrationProtocol, LabelMap,
    VoxelSize, SessionDetail, Volume, Point, Region, Modeling, ModelingProtocol
)
from qiprofile_rest_client.model.clinical import (
    Biopsy, Evaluation, Surgery, BreastSurgery, BreastPathology,
    SarcomaPathology, TNM, ModifiedBloomRichardsonGrade,
    HormoneReceptorStatus, FNCLCCGrade, NecrosisPercentValue,
    NecrosisPercentRange, necrosis_percent_as_score
)


class TestModel(object):
    """
    Basic data model test. A more complete test is found in the qiprofile_rest
    server TestSeed test suite.
    """
    def setup(self):
        connect(db='qiprofile_test')
 
    def test_subject(self):
        subject = Subject(project='QIN_Test', number=1)
        # The subject must have a collection.
        with assert_raises(ValidationError):
            subject.validate()

        subject.collection='Breast'
        subject.validate()

    def test_race(self):
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        subject.races = ['White', 'Black', 'Asian', 'AIAN', 'NHOPI']
        subject.validate()

        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        subject.races = ['Invalid']
        with assert_raises(ValidationError):
            subject.validate()

        # Races must be a list.
        subject.races = 'White'
        with assert_raises(ValidationError):
            subject.validate()

    def test_ethnicity(self):
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        subject.ethnicity = 'Non-Hispanic'
        subject.validate()

        # The ethnicity is a controlled value.
        subject.ethnicity = 'Invalid'
        with assert_raises(ValidationError):
            subject.validate()

    def test_biopsy(self):
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # The pathology.
        size = TNM.Size.parse('T3a')
        grade = ModifiedBloomRichardsonGrade(
            tubular_formation=2, nuclear_pleomorphism=1, mitotic_count=2
        )
        tnm = TNM(tumor_type='Breast', grade=grade, size=size,
                  metastasis=False, resection_boundaries=1,
                  lymphatic_vessel_invasion=False)
        estrogen=HormoneReceptorStatus(hormone='estrogen', positive=True,
                                       intensity=80)
        hormone_receptors = [estrogen]
        pathology = BreastPathology(tnm=tnm,
                                    hormone_receptors=hormone_receptors)
        # Add the encounter to the subject.
        date = datetime(2013, 1, 4)
        biopsy = Biopsy(date=date, weight=54, pathology=pathology)
        subject.encounters = [biopsy]
        # Validate the subject and embedded biopsy.
        subject.validate()

    def test_breast_surgery(self):
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # The pathology.
        size = TNM.Size.parse('T2')
        grade = ModifiedBloomRichardsonGrade(
            tubular_formation=1, nuclear_pleomorphism=1, mitotic_count=1
        )
        tnm = TNM(tumor_type='Breast', grade=grade, size=size,
                  metastasis=False, resection_boundaries=1,
                  lymphatic_vessel_invasion=False)
        pathology = BreastPathology(tnm=tnm)
        # Add the encounter to the subject.
        date = datetime(2013, 1, 4)
        surgery = BreastSurgery(date=date, weight=54, surgery_type='Lumpectomy',
                               pathology=pathology)
        subject.encounters = [surgery]
        # Validate the subject and embedded surgery.
        subject.validate()

    def test_sarcoma_surgery(self):
        subject = Subject(project='QIN_Test', collection='Sarcoma', number=1)
        # The pathology.
        size = TNM.Size.parse('T3a')
        grade = FNCLCCGrade(
            differentiation=2, necrosis_score=1, mitotic_count=1
        )
        tnm = TNM(tumor_type='Sarcoma', grade=grade, size=size,
                  metastasis=False, resection_boundaries=1,
                  lymphatic_vessel_invasion=False)
        pathology = SarcomaPathology(tnm=tnm, location='Thigh')
        # Add the encounter to the subject.
        date = datetime(2014, 6, 19)
        surgery = Surgery(date=date, weight=47, pathology=pathology)
        subject.encounters = [surgery]
        # Validate the subject and embedded surgery.
        subject.validate()

    def test_tnm_size(self):
        for value in ['T1', 'Tx', 'cT4', 'T1b', 'cT2a']:
            size = TNM.Size.parse(value)
            assert_equal(str(size), value, "The TNM parse is incorrect -"
                                           " expected %s, found %s"
                                           % (value, str(size)))

    def test_necrosis_score(self):
        fixture = {
            0: dict(integer=0,
                    value=NecrosisPercentValue(value=0),
                    range=NecrosisPercentRange(
                        start=NecrosisPercentRange.LowerBound(value=0),
                        stop=NecrosisPercentRange.UpperBound(value=1))),
            1: dict(integer=40,
                    value=NecrosisPercentValue(value=40),
                    range=NecrosisPercentRange(
                        start=NecrosisPercentRange.LowerBound(value=40),
                        stop=NecrosisPercentRange.UpperBound(value=50))),
            2: dict(integer=50,
                    value=NecrosisPercentValue(value=50),
                    range=NecrosisPercentRange(
                        start=NecrosisPercentRange.LowerBound(value=50),
                        stop=NecrosisPercentRange.UpperBound(value=60)))
        }
        for expected, inputs in fixture.iteritems():
            for in_type, in_val in inputs.iteritems():
                actual = necrosis_percent_as_score(in_val)
                assert_equal(actual, expected,
                             "The necrosis score for %s is incorrect: %d" %
                             (in_val, expected))

    def test_treatment(self):
       # TODO - add the treatment test case.
       pass

    def test_session(self):
        # The test subject.
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # The test session.
        date = datetime(2013, 1, 4)
        session = Session(date=date)
        # Validate the subject and embedded session.
        subject.encounters = [session]
        subject.validate()

    def test_add_encounter(self):
        # The test subject.
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # Add the test encounters.
        encounters = [Encounter(date=datetime(2014, m, 1)) for m in (3, 5, 7)]
        subject.add_encounter(encounters[1])
        assert_equal(subject.encounters, encounters[1:2])
        subject.add_encounter(encounters[0])
        assert_equal(subject.encounters, encounters[0:2])
        subject.add_encounter(encounters[2])
        assert_equal(subject.encounters, encounters)

    def test_modeling(self):
        # The test subject.
        subject = Subject(project='QIN_Test', collection='Breast', number=1)
        # The modeling protocol must exist.
        mdl_pcl = database.get_or_create(ModelingProtocol,
                                         dict(technique='Bolero'),
                                         input_parameters=dict(r10_val=0.7))
        # The source protocol.
        scan_pcl = database.get_or_create(ScanProtocol, dict(scan_type='T1'))
        source = Modeling.Source(scan=scan_pcl)
        # The modeling data.
        ktrans = Modeling.ParameterResult(filename='/path/to/ktrans.nii.gz')
        modeling = Modeling(protocol=mdl_pcl, source=source, resource='pk_01',
                            result=dict(ktrans=ktrans))
        # Validate the subject and embedded session modeling.
        date = datetime(2014, 3, 1)
        session = Session(date=date, modelings=[modeling])
        subject.encounters = [session]
        subject.validate()

    def test_scan(self):
        # The scan protocol.
        voxel_size = VoxelSize(width=2, depth=4, spacing=2.4)
        protocol = database.get_or_create(ScanProtocol,
                                          dict(scan_type='T1'),
                                          orientation='axial',
                                          description='T1 AX SPIN ECHO',
                                          voxel_size=voxel_size)
        # The scan.
        scan = Scan(protocol=protocol, number=1)
        # Validate the session detail and embedded scan.
        detail = SessionDetail(scans=[scan])
        detail.validate()

    def test_bolus_arrival(self):
        # The scan protocol.
        protocol = database.get_or_create(ScanProtocol, dict(scan_type='T1'))
        # The scan with a bogus bolus arrival.
        scan = Scan(protocol=protocol, number=1, bolus_arrival_index=4)
        # The detail object.
        detail = SessionDetail(scans=[scan])
        # The bolus arrival index must refer to an existing volume.
        with assert_raises(ValidationError):
            detail.validate()

        # The scan volumes.
        vol_fmt = "volume%03d.nii.gz"
        create_volume = lambda number: Volume(filename=vol_fmt % number)
        scan.volumes = [create_volume(i + 1) for i in range(32)]
        # The bolus arrival is now valid.
        detail.validate()

        # The bolus arrival index must refer to a volume.
        scan.bolus_arrival_index = 32
        with assert_raises(ValidationError):
            detail.validate()

    def test_registration(self):
        # The scan protocol.
        scan_pcl = database.get_or_create(ScanProtocol, dict(scan_type='T1'))
        # The scan.
        scan = Scan(protocol=scan_pcl, number=1)

        # The registration protocol.
        reg_params = dict(transforms=['Rigid', 'Affine', 'SyN'])
        reg_pcl = database.get_or_create(RegistrationProtocol,
                                         dict(technique='ANTS'),
                                         parameters=reg_params)
        # The registration.
        reg = Registration(protocol=reg_pcl, resource='reg_h3Fk5')

        # Validate the session detail and embedded scan registration.
        scan.registrations = [reg]
        detail = SessionDetail(scans=[scan])
        detail.scans = [scan]
        detail.validate()

    def test_roi(self):
        # The scan protocol.
        scan_pcl = database.get_or_create(ScanProtocol,
                                          dict(scan_type='T1'))
        # The scan.
        scan = Scan(protocol=scan_pcl, number=1)

        # The ROI.
        mask = '/path/to/mask.nii.gz'
        label_map = LabelMap(filename='/path/to/label_map.nii.gz',
                             color_table='/path/to/color_table.nii.gz')
        centroid = Point(x=200, y=230, z=400)
        intensity = 31
        roi = Region(mask=mask, label_map=label_map, centroid=centroid,
                     average_intensity=intensity)

        # Validate the session detail and embedded scan ROI.
        scan.rois = [roi]
        detail = SessionDetail(scans=[scan])
        detail.scans = [scan]
        detail.validate()

if __name__ == "__main__":
    import nose
    nose.main(defaultTest=__name__)



"""
The qiprofile subject Mongodb data model.
"""

import re
import mongoengine
from mongoengine import (fields, signals)
from .. import choices
from .encounter import Encounter
from .imaging import (Scan, Session)
from .clinical import Treatment


class Subject(mongoengine.Document):
    """
    The patient.
    """

    RACE_CHOICES = [('White', 'White'),
                    ('Black', 'Black or African American'),
                    ('Asian', 'Asian'),
                    ('AIAN', 'American Indian or Alaska Native'),
                    ('NHOPI', 'Native Hawaiian or Other Pacific Islander')]
    """The standard FDA race categories, in order of US incidence."""

    ETHNICITY_CHOICES = [('Hispanic' , 'Hispanic or Latino'),
                         ('Non-Hispanic' , 'Not Hispanic or Latino')]
    """The standard FDA ethnicity categories."""

    GENDER_CHOICES = ['Male', 'Female']
    """The Male and Female genders."""

    meta = dict(collection='qiprofile_subject')

    project = fields.StringField(required=True)
    """The required project name."""

    collection = fields.StringField(required=True)
    """The required collection name."""

    number = fields.IntField(required=True)
    """The required subject number."""

    birth_date = fields.DateTimeField()
    """The subject date of birth, anonymized to July 7 for deidentified data."""

    races = fields.ListField(
        fields.StringField(
            max_length=choices.max_length(RACE_CHOICES),
            choices=RACE_CHOICES))
    """The :const:`RACE_CHOICES` controlled value."""

    ethnicity = fields.StringField(
        max_length=choices.max_length(ETHNICITY_CHOICES),
        choices=ETHNICITY_CHOICES)
    """The :const:`ETHNICITY_CHOICES` controlled value."""

    gender = fields.StringField(
        max_length=choices.max_length(GENDER_CHOICES),
        choices=GENDER_CHOICES)
    """The :const:`GENDER_CHOICES` controlled value."""

    treatments = fields.ListField(field=fields.EmbeddedDocumentField(Treatment))
    """The list of subject treatments."""

    encounters = fields.ListField(field=fields.EmbeddedDocumentField(Encounter))
    """The list of subject encounters in temporal order."""

    @property
    def sessions(self):
        """
        :return: the :class:`qiprofile_rest_client.imaging.Session`
            encounters
        """
        return (enc for enc in self.encounters if self._is_session(enc))

    def add_encounter(self, encounter):
        """
        Inserts the given encounter to this :class:`Subject` encounters
        list in temporal order by encounter date.
        """
        index = next((i for i, enc in enumerate(self.encounters)
                      if enc.date > encounter.date), len(self.encounters))
        # Add the new encounter to the subject encounters list.
        self.encounters.insert(index, encounter)

    @property
    def clinical_encounters(self):
        """
        :return: the non-:class:`qiprofile_rest_client.imaging.Session`
            encounters
        """
        return (enc for enc in self.encounters if not self._is_session(enc))

    def pre_delete(cls, sender, document, **kwargs):
        """Cascade delete the subject's sessions."""

        for sess in self.sessions:
            sess.delete()

    def _is_session(self, encounter):
        return isinstance(encounter, Session)

    def __str__(self):
        return ("%s %s Subject %d" %
                (self.project, self.collection, self.number))

signals.pre_delete.connect(Subject.pre_delete, sender=Subject)



"""
The qiprofile-rest unit tests.

:Note: These tests require an active qiprofile REST server.
  The tests create and delete a ``qiprofile_test`` database.
  A more thorough model test is provided by the ``qiprofile-rest``
  server project's ``test_seed.py`` unit test.
"""



"""Mongo Engine interaction utilities."""

def get_or_create(klass, pk, **non_pk):
    """
    This function stands in for the Mongo Engine ``get_or_create``
    collection method which was deprecated in mongoengine v0.8.0
    and dropped in mongoengine v0.10.0, since MongoDB does not
    support transactions.
    
    If there is an object of the given Mongo Engine data model
    class which matches the primary key, then that object is
    returned. Otherwise, a new object is created with the content
    prescribed by both the primary and non-primary parameters.
    
    The create step is an upsert, i.e. a new object is created only
    if it does not yet exist. The upsert allows for the small
    possibility that an object is created after the fetch attempt
    but before the create attempt. In that situation, the existing
    object non-key content is modified and the modified object is
    returned.
    
    :Note: The idiom used in this function modifies the solution
    proposed in http://stackoverflow.com/questions/25846462/mongoengine-replacing-get-or-create-with-upsert-update-one/25863633#25863633.
    That StackOverflow work-around returns the following error:
    
        ValueError: update only works with $ operators

    The work-around to the StackOverflow work-around is to use
    call *update* rather than *modify*.

    :param klass: the Mongo Engine data model class
    :param pk: the primary key {attribute: value} dictionary
    :param non_pk: the non-key {attribute: value} dictionary
    :return: the existing or new object
    """
    try:
        return klass.objects.get(**pk)
    except klass.DoesNotExist:
        mod_opts = {'set__' + attr: val for attr, val in non_pk.iteritems()}
        return klass.objects(**pk).update_one(upsert=True, **mod_opts)



"""This facade helper module provides a qiprofile REST facade."""

import os
import re
from qiutil.logging import logger
from .model.subject import Subject
from .model.imaging import (Session, Modeling)

SBJ_REGEX = re.compile('^(\w+?)(\d+)$')

SESS_REGEX = re.compile('^Session(\d+)$')


def find_subject(project, subject, create=False):
    """
    Finds the given QuIP subject. If it doesn't exists
    and the *create* flag is set, then the subject is created.

    :param project: the project name
    :param subject: the subject name
    :param create: flag indicating whether to create the subject
    :return: the Subject instance, or None if not found
    """
    # The collection name and subject number.
    collection, sbj_nbr_str = SBJ_REGEX.match(subject).groups()
    sbj_nbr = int(sbj_nbr_str)

    # Fetch the subject.
    try:
        sbj = Subject.objects.get(project__name=project,
                                  collection__name=collection,
                                  number=sbj_nbr)
    except Subject.DoesNotExist:
        if create:
            sbj = Subject(project=project, collection=collection, number=sbj_nbr)
            sbj.save()
        else:
            return

    return sbj


def find_session(project, subject, session, create=False):
    """
    Finds the given QuIP subject. If it doesn't exists
    and the *create* flag is set, then the session is created.

    :param project: the project name
    :param subject: the subject name
    :param session: the session name
    :param create: flag indicating whether to create the subject
    :return: the Session instance, or None if not found
    """
    # The session number.
    sess_nbr_str = SESS_REGEX.match(session).group(1)
    sess_nbr = int(sess_nbr_str)

    # Fetch the subject.
    sbj = find_subject(project, subject, create)
    if not sbj:
        return
    try:
        sess = Session.objects.get(subject=sbj, number=sess_nbr)
    except Subject.DoesNotExist:
        if create:
            sess = Session(subject=subject, number=sess_nbr)
            sess.save()
        else:
            return

    return sess


def save_subject(project, subject, **opts):
    """
    Creates the QuIP subject, if necessary, and updates
    the given subject fields.

    :param project: the project name
    :param subject: the subject name
    :param opts: the ``qiprofile.models.Subject`` non-identifier
        field name-value parameters
    :return: the Subject instance
    """
    # The qiprofile collection and subject number.
    collection, sbj_nbr_str = SBJ_REGEX.match(subject).groups()
    sbj_nbr = int(sbj_nbr_str)

    # Make the subject.
    sbj = Subject(project=project, collection=collection, number=sbj_nbr)
    for attr, value in opts.iteritems():
        setattr(sbj, attr, value)

    # Save the subject.
    logger(__name__).debug("Saving %s Subject %s..." % (project, subject))
    sbj.save()
    logger(__name__).debug("Saved %s Subject %s." % (project, subject))


def save_session(project, subject, session, **opts):
    """
    :param project: the project name
    :param subject: the subject name
    :param session: the session name
    :param opts: the ``qiprofile.models.Session`` non-identifier field
        name-value parameters
    :return: the Session instance
    """
    # The session number.
    sess_nbr_str = SESS_REGEX.match(session).group(1)
    sess_nbr = int(sess_nbr_str)

    # The subject.
    sbj = find_subject(project, subject, create=True)

    # Make the session.
    sess = Session(subject=sbj, number=sess_nbr, **opts)

    # Save the session.
    logger(__name__).debug("Saving %s Subject %s Session %s..." %
                           (project, subject, session))
    sess.save()
    logger(__name__).debug("Saved %s Subject %s Session %s." %
                           (project, subject, session))


def save_modeling(project, subject, session, assessor, **opts):
    """
    :param project: the project id
    :param subject: the subject name
    :param session: the session name
    :param assessor: the XNAT assessor name
    :param opts: the ``qiprofile.models.Modeling`` non-identifier field
        name-value parameters
    :return: the Modeling instance
    """
    # The session.
    sess = find_session(project, subject, session, create=True)

    # The modeling object.
    mdl = Modeling(session=sess, name=assessor, **opts)
    mdl.save()



"""
The qiprofile encounter abstract MongoDB data model.
"""

import mongoengine
from mongoengine import (fields, ValidationError)


class Encounter(mongoengine.EmbeddedDocument):
    """The patient clinical encounter, e.g. biopsy or imaging session."""

    meta = dict(allow_inheritance=True)

    date = fields.DateTimeField(required=True)
    """The encounter date."""

    weight = fields.IntField()
    """The patient weight in kilograms."""
